/*

*/

import java.util.Arrays;

/**
   {@code ArrayList<E>} implements a generic list of values that can grow and shrink.
   <p>
   The public interface of this list ADT has the following methods.
   <p>
      int     size()                       <br>
      void    add(int index, E e)          <br>
      E       remove(int index)            <br>
      boolean remove(E e)                  <br>
      void    set(int index, E e)          <br>
      E       get(int index)               <br>
      void    clear()                      <br>
      boolean isEmpty()                    <br>
      boolean contains(E e)                <br>
      int     indexOf(E e)                 <br>
      int     lastIndexOf(E e)             <br>
*/
public class ArrayList<E> extends List<E>
{
   private E[] items;     // partially full array holding the elements
   private int itemCount; // how many elements are in the list

   /**
      Default constructor creates an empty list with a backing array of size eight.
   */
   public ArrayList()
   {
      @SuppressWarnings("unchecked")
      final E[] items = (E[])new Object[8];
      this.items = items;

      itemCount = 0;
   }


   /**
      Returns the number of items currently in this {@code List}.

      @return the number of items in this {@code List}
   */
   @Override
   public int size()
   {
      return itemCount;
   }


   /**
      Return the item at the given {@code index} in this {@code List}.

      @param index  location in this {@code List} of the item to retrieve
      @return a reference to the item in this {@code List} at location {@code index}
   */
   @Override
   public E get(int index)
   {
    //if ( index < 0 || index >= size() )  // self encapsulation
      if ( index < 0 || index >= itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      return items[index];
   }


   /**
      Set the value of the {@code List} item at the given {@code index}.

      @param index  location in this {@code List} of the item to replace
      @param item   reference to the item to place in this {@code List}
   */
   @Override
   public void set(int index, E item)
   {
      if ( index < 0 || index >= itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      items[index] = item;
   }


   /**
      Insert the item at the given {@code index} in this {@code List}.
      The {@code List} size increases by one.
      <p>
      For this method, the index is valid if 0 <= index <= size().
      <p>
      If the index is equal to the size of the list, that means that we are
      adding an elelment to the end of the list. In particular, if the list
      is empty, we can add an item at index 0, even though 0 is not (yet)
      a valid index into the list.
      <p>
      Items following the added item are shifted up one space in the backing array.

      @param index  location in this {@code List} for the inserted {@code item}
      @param item   reference to the item to be added to this {@code List}
   */
   @Override
   public void add(int index, E item)
   {
      if ( index < 0 || index > itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      if (itemCount == items.length)
      {
         items = Arrays.copyOf( items,    2*items.length ); // good for small arrays
       //items = Arrays.copyOf( items, 1000+items.length ); // good for large arrays
       //items = Arrays.copyOf( items,    1+items.length ); // bad idea
      }

      for (int i = itemCount; i > index; --i)
      {
         items[i] = items[i - 1];
      }

      items[index] = item;

      ++itemCount;
   }


   /**
      Remove the item at the given {@code index} in this {@code List}.
      The {@code List} size decreases by one.

      Items following the removed item are moved down one space in the backing array.

      @param index  location in this {@code List} of the item to remove
      @return a reference to the item that was removed
   */
   @Override
   public E remove(int index)
   {
      if ( index < 0 || index >= itemCount )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      E removedItem = items[index];

      for (int i = index + 1; i < itemCount; ++i)
      {
         items[i - 1] = items[i];
      }

      --itemCount;

      return removedItem;
   }


   /**
      Remove all the items in this {@code List}.
      The {@code List} size becomes zero.
   */
   @Override
   public void clear()
   {
      for (int i = 0; i < itemCount; ++i)
      {
         items[i] = null;
      }

      itemCount = 0;
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the first occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the first index of {@code item} if it is found, -1 otherwise
   */
   @Override
   public int indexOf(E item)
   {
      for (int i = 0; i < itemCount; ++i)
      {
         if ( item == null ? items[i] == null : item.equals(items[i]) )
         {
            return i;
         }
      }

      return -1;
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the last occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the last index of {@code item} if it is found, -1 otherwise
   */
   @Override
   public int lastIndexOf(E item)
   {
      for (int i = itemCount - 1; i >= 0; --i)
      {
         if ( item == null ? items[i] == null : item.equals(items[i]) )
         {
            return i;
         }
      }

      return -1;
   }
}
