/*

*/



/**
   {@code LinkedList<E>} implements a generic list of values that can grow and shrink.
   <p>
   The public interface of this list ADT has the following methods.
   <p>
      int     size()                       <br>
      void    add(int index, E e)          <br>
      E       remove(int index)            <br>
      boolean remove(E e)                  <br>
      void    set(int index, E e)          <br>
      E       get(int index)               <br>
      void    clear()                      <br>
      boolean isEmpty()                    <br>
      boolean contains(E e)                <br>
      int     indexOf(E e)                 <br>
      int     lastIndexOf(E e)             <br>
*/
public class LinkedList<E> extends List<E>
{
   private Node<E> head;
   private int length;

   /**
      Default constructor creates an empty list.
   */
   public LinkedList()
   {
      this.head = null;
      this.length = 0;
   }


   /**
      Returns the number of items currently in this {@code List}.

      @return the number of items in this {@code List}
   */
   @Override
   public int size()
   {
      return this.length;
   }


   /**
      Return the item at the given {@code index} in this {@code List}.

      @param index  location in this {@code List} of the item to retrieve
      @return a reference to the item in this {@code List} at location {@code index}
   */
   @Override
   public E get(int index)
   {
    //if ( index < 0 || index >= size() )  // self encapsulation
      if ( index < 0 || index >= length )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      Node<E> cursor = head;
      for (int i = 0; i < index; ++i)
      {
         cursor = cursor.getNext();
      }
      return cursor.getData();
   }


   /**
      Set the value of the {@code List} item at the given {@code index}.

      @param index  location in this {@code List} of the item to replace
      @param item   reference to the item to place in this {@code List}
   */
   @Override
   public void set(int index, E item)
   {
      if ( index < 0 || index >= length )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      Node<E> cursor = head;
      for (int i = 0; i < index; ++i)
      {
         cursor = cursor.getNext();
      }
      cursor.setData(item);
   }


   /**
      Insert the item at the given {@code index} in this {@code List}.
      The {@code List} size increases by one.
      <p>
      For this method, the index is valid if 0 <= index <= size().
      <p>
      If the index is equal to the size of the list, that means that we are
      adding an elelment to the end of the list. In particular, if the list
      is empty, we can add an item at index 0, even though 0 is not (yet)
      a valid index into the list.

      @param index  location in this {@code List} for the inserted {@code item}
      @param item   reference to the item to be added to this {@code List}
   */
   @Override
   public void add(int index, E item)
   {
      if ( index < 0 || index > length )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      if (0 == index)
      {
         head = new Node<E>(item, head);
      }
      else
      {
         Node<E> cursor = head;
         for (int i = 0; i < index - 1; ++i)
         {
            cursor = cursor.getNext();
         }
         cursor.setNext( new Node<E>(item, cursor.getNext()) );
      }

      ++this.length;
   }


   /**
      Remove the item at the given {@code index} in this {@code List}.
      The {@code List} size decreases by one.

      @param index  location in this {@code List} of the item to remove
      @return a reference to the item that was removed
   */
   @Override
   public E remove(int index)
   {
      if ( index < 0 || index >= length )
      {
         throw new IndexOutOfBoundsException("Illegal List index, " + index);
      }

      final E removedItem;
      if (0 == index)
      {
         removedItem = head.getData();
         head = head.getNext();
      }
      else
      {
         Node<E> cursor = head;
         for (int i = 0; i < index - 1; ++i)
         {
            cursor = cursor.getNext();
         }
         removedItem = cursor.getNext().getData();
         cursor.setNext( cursor.getNext().getNext() );
      }
      --this.length;
      return removedItem;
   }


   /**
      Remove all the items in this {@code List}.
      The {@code List} size becomes zero.
   */
   @Override
   public void clear()
   {
      this.head = null; // the nodes in this list will be garbage collected
      this.length = 0;
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the first occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the first index of {@code item} if it is found, -1 otherwise
   */
   @Override
   public int indexOf(E item)
   {
      Node<E> cursor = head;
      int index = 0;
      while (null != cursor)
      {
         if ( cursor.getData().equals(item) )
         {
            return index; // found
         }
         ++index;
         cursor = cursor.getNext();
      }
      return -1; // not found
   }


   /**
      Search this {@code List} for the given {@code item}.
      Return the index of the last occurrence of the
      {@code item} if it is found and return {@code -1}
      if the item is not found.

      @param item  reference to the item to search for in this {@code List}
      @return the last index of {@code item} if it is found, -1 otherwise
   */
   @Override
   public int lastIndexOf(E item)
   {
      Node<E> cursor = head;
      int i = 0;
      int index = -1;
      while (null != cursor)
      {
         if ( cursor.getData().equals(item) )
         {
            index = i; // found
         }
         ++i;
         cursor = cursor.getNext();
      }
      return index;
   }
}
